/*
 * The Node class contains information about a specific state of the game.
 * The information consists off:
 * -Depth of the node
 * -Parent of the node
 * -Childs of the node
 * -a value, depending on the state of the game
 * -the current state of the game
 * 
 * 
 * The node on the top of the tree contains the current state of the playing field.
 * Every child of this node consists of the initial playing field and a unique
 * new disc placed. This is done using an argument in the constructor.
 */
package AI;

import Game.Disc;
import Game.Player;
import Game.PlayingField;

/**
 * @author Ruben Jonk/Romain Emens
 */
public class Node {

    //variables
    //depth, Parent, Childs
    private int depth;  //stores the depth of the node. depth top node == 0
    private Node parent;        //stores the parent node
    //public Node[] child;      //stores all the childs this node might have
    public int nchilds;        //stores the amount of childs this node currently has
    //playing field and heuristic value
    PlayingField field; //instance containing the state of the field
    public int heurval; //the heuristic value of this node
    private int maxcol = 0;
    private int fwidth;
    private int fheight;
    private Player turn;

    /**
     * Make a new node with a field and a new move.
     * 
     * @param depth The depth of the node.
     * @param par The parent of the node.
     * @param afield The field of the parent or the initial field if the node is the root node.
     * @param move The move to perform on the field. Move <0 indicates no move done.
     */
    Node(int depth, Node par, PlayingField afield, int move) {
        this.depth = depth;
        parent = par;
        nchilds = 0;


        field = afield.clone();

        maxcol = field.getheight() - 1;
        fheight = maxcol;
        fwidth = field.getWidth() - 1;
        turn = field.getTurn();


        if (field.isColumnFull(move)) {
            if (depth % 2 == 1) {
                heurval = -1001;
            } else {
                heurval = 1001;
            }
        }


        if (move > -1) {
            field.putDisc(move, turn);
        }
        //child = new Node[field.getWidth()];


        if (field.checkWin() != null) {
            if (depth % 2 == 1) {
                heurval = 999;
            } else {
                heurval = -999;
            }
        }
    }

    /**
     * Set a new child. The new child will do a new move to the furthermost left column
     * where there has not been a move before when creating a child from the same parent.
     * @return A new child based on the parent.
     */
    public Node setChild() {
        Node nNode = new Node(depth - 1, this, field, nchilds);
        //child[nchilds] = nNode;
        nchilds++;
        return nNode;
    }

    /**
     * Clone a node.
     * @return Cloned node.
     */
    @Override
    public Node clone() {
        Node n = new Node(depth, parent, field, -1);
        return n;
    }

    /**
     * Calculates the heuristic value of the node.
     * 
     * The method first checks for a winning situation, a full column or a full 
     * playing field. Afterwards various checks are performed to check for certain 
     * sequences in the playing field.
     * @return The heuristic value of the node.
     */
    public int calHeuristicValue() {
        if (heurval == 999) {
            heurval = 980 + depth;
            return heurval;
        }

        if (heurval == -999) {
            heurval = -980 - depth;
            return heurval;
        }
        if (heurval == 1001 || heurval == -1001) {
            return heurval;
        }

        if (field.isFull()) {
            heurval = 0;
            return heurval;
        }


        maxcol = 0;
        for (int i = 0; i <= fwidth; i++) {
            if (field.discsincol[i] > maxcol) {
                maxcol = field.discsincol[i];
            }
        }
        maxcol--;


        if (topleftbotright_2_c()) {           // top left to bottem right, C
            heurval = 920;
            Node n = this.clone();
            while (n.parent != null) {
                if (n.parent.topleftbotright_2_c()) {
                    heurval++;
                } else {
                    break;
                }
                n = n.parent;
            }
            return heurval;
        }

        if (topleftbotright_2_h()) {             // top left to bottem right, H
            heurval = -920;
            Node n = this;
            while (n.parent != null) {
                if (n.parent.topleftbotright_2_h()) {
                    heurval--;
                } else {
                    break;
                }
                n = n.parent;
            }
            return heurval;
        }


        if (toprightbotleft_2_c()) {           // top right to bottem left, C
            heurval = 920;
            Node n = this;
            while (n.parent != null) {
                if (n.parent.toprightbotleft_2_c()) {
                    heurval++;
                } else {
                    break;
                }
                n = n.parent;
            }
            return heurval;
        }

        if (toprightbotleft_2_h()) {             // top right to bottem left, H
            heurval = -920;
            Node n = this;
            while (n.parent != null) {
                if (n.parent.toprightbotleft_2_h()) {
                    heurval--;
                } else {
                    break;
                }
                n = n.parent;
            }
            return heurval;
        }

        //int[] seqlen = new int[2];
        //horizontal_sequence(seqlen);
        //if (seqlen[0] >= 2) return heurval = 900;
        //if (seqlen[1] >= 2) return heurval = -900;
        //if (seqlen[0] >= 1) return heurval = 820;
        //if (seqlen[1] >= 1) return heurval = -820;

        if (horizontal_2_c()) {
            //System.out.println("c seq=" + Integer.toString(seqlen[0]) + "," + Integer.toString(seqlen[1]));
            heurval = 900;
            Node n = this;
            while (n.parent != null) {
                if (n.parent.horizontal_2_c()) {
                    heurval++;
                } else {
                    break;
                }
                n = n.parent;
            }
            return heurval;
        }

        if (horizontal_2_h()) {
            //System.out.println("h seq=" + Integer.toString(seqlen[0]) + "," + Integer.toString(seqlen[1]));
            heurval = -900;
            Node n = this;
            while (n.parent != null) {
                if (n.parent.horizontal_2_h()) {
                    heurval--;
                } else {
                    break;
                }
                n = n.parent;
            }
            return heurval;
        }

        if (toprightbotleft_1g_c()) {
            heurval = 880;
            Node n = this;
            while (n.parent != null) {
                if (n.parent.toprightbotleft_1g_c()) {
                    heurval++;
                } else {
                    break;
                }
                n = n.parent;
            }
            return heurval;
        }


        if (topleftbotright_1g_c()) {
            heurval = 880;
            Node n = this;
            while (n.parent != null) {
                if (n.parent.topleftbotright_1g_c()) {
                    heurval++;
                } else {
                    break;
                }
                n = n.parent;
            }
            return heurval;
        }

        if (topleftbotright_1g_h()) {
            heurval = -880;
            Node n = this;
            while (n.parent != null) {
                if (n.parent.topleftbotright_1g_h()) {
                    heurval--;
                } else {
                    break;
                }
                n = n.parent;
            }
            return heurval;
        }

        if (toprightbotleft_1g_h()) {
            heurval = -880;
            Node n = this;
            while (n.parent != null) {
                if (n.parent.toprightbotleft_1g_h()) {
                    heurval--;
                } else {
                    break;
                }
                n = n.parent;
            }
            return heurval;
        }


        if (toprightbotleft_1_c()) {
            heurval = 860;
            Node n = this;
            while (n.parent != null) {
                if (n.parent.toprightbotleft_1_c()) {
                    heurval++;
                } else {
                    break;
                }
                n = n.parent;
            }
            return heurval;
        }


        if (toprightbotleft_1_c()) {
            heurval = 860;
            Node n = this;
            while (n.parent != null) {
                if (n.parent.toprightbotleft_1_c()) {
                    heurval++;
                } else {
                    break;
                }
                n = n.parent;
            }
            return heurval;
        }

        if (toprightbotleft_1_h()) {
            heurval = -860;
            Node n = this;
            while (n.parent != null) {
                if (n.parent.toprightbotleft_1_h()) {
                    heurval--;
                } else {
                    break;
                }
                n = n.parent;
            }
            return heurval;
        }

        if (toprightbotleft_1_h()) {
            heurval = -860;
            Node n = this;
            while (n.parent != null) {
                if (n.parent.toprightbotleft_1_h()) {
                    heurval--;
                } else {
                    break;
                }
                n = n.parent;
            }
            return heurval;
        }

        if (horizontal_1gap_h()) {
            heurval = -840;
            Node n = this;
            while (n.parent != null) {
                if (n.parent.horizontal_1gap_h()) {
                    heurval--;
                } else {
                    break;
                }
                n = n.parent;
            }
            return heurval;
        }


        if (horizontal_1_c()) {
            heurval = 820;
            Node n = this;
            while (n.parent != null) {
                if (n.parent.horizontal_1_c()) {
                    heurval++;
                } else {
                    break;
                }
                n = n.parent;
            }
            return heurval;
        }

        if (horizontal_1_h()) {
            heurval = -820;
            Node n = this;
            while (n.parent != null) {
                if (n.parent.horizontal_1_h()) {
                    heurval--;
                } else {
                    break;
                }
                n = n.parent;
            }
            return heurval;
        }

        if (vertical_1_c()) {
            heurval = 800;
            Node n = this;
            while (n.parent != null) {
                if (n.parent.vertical_1_c()) {
                    heurval++;
                } else {
                    break;
                }
                n = n.parent;
            }
            return heurval;
        }


        if (vertical_1_h()) {
            heurval = -800;
            Node n = this;
            while (n.parent != null) {
                if (n.parent.vertical_1_h()) {
                    heurval--;
                } else {
                    break;
                }
                n = n.parent;
            }
            return heurval;
        }

        if (vertical_2_c()) {
            heurval = 50;
            return heurval;
        }

        if (vertical_2_h()) {
            heurval = -50;
            return heurval;
        }

        heurval = horizontaltwo_2gap_c();


        return heurval; //default implementation
    }

    //Functions for heuristic value  
    /**
     * Function to check topleft to botright diagonal with 2 chances on both ends for max player
     * @return True if sequence is found, false otherwise.
     */
    private boolean topleftbotright_2_c() {


        // Topleft to Bottomright diagonal
        // two empty places at top and bottom
        //   [_]
        //     [X]
        //       [x]
        //         [x]'           (' = start point van r)
        //            [_]
        // Computer player 

        if (fwidth < 4 || maxcol < 4) {
            return false;
        } else { // fwidth = field.getWidth - 1, maxcol = rownumber that stores the highest disk

            int cons = 0;
            int r = 0;
            int ccol = 3;
            int crow = 1;
            while (ccol < fwidth) { // fwidth = field.getWidth - 1
                crow = 1;
                while (crow <= maxcol - 3) {
                    cons = 0;
                    for (r = 0; r <= 2; r++) {
                        if (field.getDiscAt(ccol - r, crow + r) != null // |_
                                && field.getDiscAt(ccol - r, crow + r).getOwner() == turn) {
                            cons++;
                        } else {
                            cons = 0;
                        }
                        if (cons == 3
                                && field.getDiscAt(ccol - r - 1, crow + r + 1) == null
                                && field.getDiscAt(ccol + 1, crow - 1) == null) {
                            return true;
                        }
                    }
                    crow++;
                }
                ccol++;
            }
            return false;
        }
    }

    /**
     * Function to check topleft to botright diagonal with 2 chances on both ends for min player
     * @return True if sequence is found, false otherwise.
     */
    private boolean topleftbotright_2_h() {

        // Topleft to Bottomright diagonal
        // two empty places
        //   [_]
        //     [X]
        //       [x]
        //         [x]'           (' = start point of r)
        //            [_]
        // human player   

        if (fwidth < 4 || maxcol < 4) {
            return false;
        } else {

            int cons = 0;
            int r = 0;
            int ccol = 3;
            int crow = 1;
            while (ccol < fwidth) {
                crow = 1;
                while (crow <= maxcol - 3) {
                    cons = 0;
                    for (r = 0; r <= 2; r++) {
                        if (field.getDiscAt(ccol - r, crow + r) != null // |_
                                && field.getDiscAt(ccol - r, crow + r).getOwner() != turn) {
                            cons++;
                        } else {
                            cons = 0;
                        }
                        if (cons == 3
                                && field.getDiscAt(ccol - r - 1, crow + r + 1) == null
                                && field.getDiscAt(ccol + 1, crow - 1) == null) {
                            return true;
                        }
                    }
                    crow++;
                }
                ccol++;
            }
            return false;
        }
    }

    /**
     * Function to check topright to botleft diagonal with 2 chances on both ends for max player
     * @return True if sequence is found, false otherwise.
     */
    private boolean toprightbotleft_2_c() {

        // Bottomleft to topright diagonal
        // two empty places
        //              [_]
        //          [X]'
        //       [x]
        //    [x]
        // [_]
        // Computer player
        if (fwidth < 4 || maxcol < 4) {
            return false;
        } else {
            int cons = 0;
            int r = 0;
            int ccol = 3;
            int crow = 3;

            while (ccol < fwidth) {
                crow = 3;
                while (crow <= maxcol - 1) {
                    cons = 0;
                    for (r = 0; r <= 2; r++) {
                        if (field.getDiscAt(ccol - r, crow - r) != null //|-
                                && field.getDiscAt(ccol - r, crow - r).getOwner() == turn) {
                            cons++;
                        } else {
                            cons = 0;
                        }
                        if (cons == 3
                                && field.getDiscAt(ccol - r - 1, crow - r - 1) == null
                                && field.getDiscAt(ccol + 1, crow + 1) == null) {
                            return true;
                        }
                    }
                    crow++;
                }
                ccol++;
            }
            return false;
        }
    }

    /**
     * Function to check topright to botleft diagonal with 2 chances on both ends for min player
     * @return True if sequence is found, false otherwise.
     */
    private boolean toprightbotleft_2_h() {

        // Bottomleft to topright diagonal
        // two empty places
        //              [_]
        //          [X]'
        //       [x]
        //    [x]
        // [_]
        // Computer player 
        if (fwidth < 4 || maxcol < 4) {
            return false;
        } else {
            int cons = 0;
            int r = 0;
            int ccol = 3;
            int crow = 3;

            while (ccol < fwidth) {
                crow = 3;
                while (crow <= maxcol - 1) {
                    cons = 0;
                    for (r = 0; r <= 2; r++) {
                        if (field.getDiscAt(ccol - r, crow - r) != null
                                && field.getDiscAt(ccol - r, crow - r).getOwner() != turn) {
                            cons++;
                        } else {
                            cons = 0;
                        }
                        if (cons == 3
                                && field.getDiscAt(ccol - r - 1, crow - r - 1) == null
                                && field.getDiscAt(ccol + 1, crow + 1) == null) {
                            return true;
                        }
                    }
                    crow++;
                }
                ccol++;
            }
            return false;
        }
    }

    /**
     * Function to check horizontal with 2 chances on both ends for max player
     * @return True if sequence is found, false otherwise.
     */
    private boolean horizontal_2_c() {

        // Horizontal three in a row, two empty sides, 
        // [_] [x]' [x] [x] [_]
        // Computer player
        if (fwidth < 4) {
            return false;
        } else {
            int col = 1;
            int row = 0;
            int cons = 0;
            while (row <= maxcol) {
                Disc d = field.getDiscAt(col, row);
                if (d != null && d.getOwner() == turn) {
                    cons++;
                } else {
                    cons = 0;
                }
                if (cons == 3 && field.getDiscAt(col - 3, row) == null
                        && field.getDiscAt(col + 1, row) == null) {

                    return true;
                }
                col++;
                if (col == fwidth + 1) {
                    row++;
                    col = 1;
                    cons = 0;
                }
            }
            return false;
        }
    }

    /**
     * Function to check horizontal with 2 chances on both ends for min player
     * @return True if sequence is found, false otherwise.
     */
    private boolean horizontal_2_h() {
        // Horizontal three in a row, two empty sides H
        // [_] [x]' [x] [x] [_]
        // Human player
        if (fwidth < 4) {
            return false;
        } else {
            int col = 1;
            int row = 0;
            int cons = 0;

            while (row <= maxcol) {
                Disc d = field.getDiscAt(col, row);
                if (d != null && d.getOwner() != turn) {
                    cons++;
                } else {
                    cons = 0;
                }
                if (cons == 3 && field.getDiscAt(col - 3, row) == null
                        && field.getDiscAt(col + 1, row) == null) {
                    return true;
                }
                col++;
                if (col == fwidth + 1) {
                    row++;
                    col = 1;
                    cons = 0;
                }
            }
            return false;
        }
    }

    /**
     * Function to check topright to botleft diagonal with 1 chance on either end for max player
     * @return True if sequence is found, false otherwise.
     */
    private boolean toprightbotleft_1_c() {

        // Top right to bottem left diagonal, with one gap at end or front
        //          [_]                 [x]'
        //       [x]'                [x]
        //    [x]         or      [x]
        // [x]                 [_]
        // computer player
        int cons = 0;
        int r = 0;
        int ccol = 2;
        int crow = 2;

        while (ccol <= fwidth) {
            crow = 2;
            while (crow <= maxcol) {
                cons = 0;
                for (r = 0; r <= 2; r++) {
                    if (field.getDiscAt(ccol - r, crow - r) != null //|-
                            && field.getDiscAt(ccol - r, crow - r).getOwner() == turn) {
                        cons++;
                    } else {
                        cons = 0;
                    }
                    if (cons == 3) {

//                        check bottomleft
                        if (ccol - 2 > 0 && crow > 2 && field.getDiscAt(ccol - 3, crow - 3) == null) {
                            return true;
                        }
//                        check topright
                        if (ccol < fwidth && crow < fheight && field.getDiscAt(ccol + 1, crow + 1) == null) {
                            return true;
                        }

                    }
                }
                crow++;
            }
            ccol++;
        }
        return false;
    }

    /**
     * Function to check topright to botleft diagonal with 1 chance on either end for min player
     * @return True if sequence is found, false otherwise.
     */
    private boolean toprightbotleft_1_h() {
        // Top right to bottem left diagonal, with one gap at end or frond    
        //          [_]                 [x]'
        //       [x]'                [x]
        //    [x]         of      [x]
        // [x]                 [_]
        // human player
        int cons = 0;
        int r = 0;
        int ccol = 2;
        int crow = 2;

        while (ccol < fwidth) {
            crow = 2;
            while (crow < maxcol) {
                cons = 0;
                for (r = 0; r <= 2; r++) {
                    if (field.getDiscAt(ccol - r, crow - r) != null //|-
                            && field.getDiscAt(ccol - r, crow - r).getOwner() != turn) {
                        cons++;
                    } else {
                        cons = 0;
                    }
                    if (cons == 3) {
//                        check bottomleft
                        if (ccol - 2 > 0 && crow > 2 && field.getDiscAt(ccol - 3, crow - 3) == null) {
                            return true;
                        }
//                        check topright
                        if (ccol < fwidth && crow < fheight && field.getDiscAt(ccol + 1, crow + 1) == null) {
                            return true;
                        }
                    }
                }
                crow++;
            }
            ccol++;
        }


        return false;
    }

    /**
     * Function to check topleft to botright diagonal with 1 chance on either end for max player
     * @return True if sequence is found, false otherwise.
     */
    private boolean topleftbotright_1_c() {
        // Top left to bottem right diagonal, with one gap at end or front
        //   [_]                   [x]   
        //     [X]                    [x]
        //       [x]                     [x]'
        //         [x]'   of                [_]

        int cons = 0;
        int r = 0;
        int ccol = 2;
        int crow = 1;
        while (ccol < fwidth) {
            crow = 1;
            while (crow <= maxcol - 2) {
                cons = 0;
                for (r = 0; r <= 2; r++) {
                    if (field.getDiscAt(ccol - r, crow + r) != null // |_
                            && field.getDiscAt(ccol - r, crow + r).getOwner() == turn) {
                        cons++;
                    } else {
                        cons = 0;
                    }
                    if (cons == 3) {
//                        check bottomright
                        if (ccol < fwidth && crow > 0 && field.getDiscAt(ccol + 1, crow - 1) == null) {
                            return true;
                        }
//                        check topleft
                        if (ccol > 2 && crow < fheight - 2 && field.getDiscAt(ccol - 3, crow + 3) == null) {
                            return true;
                        }
                    }
                }
                crow++;
            }
            ccol++;
        }
        return false;
    }

    /**
     * Function to check topleft to botright diagonal with 1 chance on either end for min player
     * @return True if sequence is found, false otherwise.
     */
    private boolean topleftbotright_1_h() {

        // Top left to bottem right diagonal, with one gap at end or front
        //   [_]                   [x]   
        //     [X]                    [x]
        //       [x]                     [x]'
        //         [x]'   of                [_]

        int cons = 0;
        int r = 0;
        int ccol = 2;
        int crow = 1;
        while (ccol < fwidth) {
            crow = 1;
            while (crow <= maxcol - 2) {
                cons = 0;
                for (r = 0; r <= 2; r++) {
                    if (field.getDiscAt(ccol - r, crow + r) != null
                            && field.getDiscAt(ccol - r, crow + r).getOwner() != turn) {
                        cons++;
                    } else {
                        cons = 0;
                    }
                    if (cons == 3) {
//                        check bottomright
                        if (ccol < fwidth && crow > 0 && field.getDiscAt(ccol + 1, crow - 1) == null) {
                            return true;
                        }
//                        check topleft
                        if (ccol > 2 && crow < fheight - 2 && field.getDiscAt(ccol - 3, crow + 3) == null) {
                            return true;
                        }
                    }
                }
                crow++;
            }
            ccol++;
        }
        return false;
    }

    /**
     * Function to check topright to botleft diagonal with 1 chance in the center for max player
     * @return True if sequence is found, false otherwise.
     */
    private boolean toprightbotleft_1g_c() {

        // Top right to bottem left diagonal, with one gap in between                   
        //          [x]                 [x]
        //       [_]                 [x]
        //    [x]         of      [_]
        // [x]'                 [x]'
        int ccol = 0;
        int crow = 0;
        Disc d;

        while (ccol < fwidth - 2) {
            crow = 0;
            while (crow <= maxcol - 3) {
                d = field.getDiscAt(ccol, crow);
                if (d != null && d.getOwner() == turn) {
                    d = field.getDiscAt(ccol + 3, crow + 3);
                    if (d != null && d.getOwner() == turn) {
                        d = field.getDiscAt(ccol + 1, crow + 1);
                        Disc e = field.getDiscAt(ccol + 2, crow + 2);

                        if ((d != null && d.getOwner() == turn && e == null)
                                || (d == null && e != null && e.getOwner() == turn)) {
                            return true;
                        }
                    }
                }
                crow++;
            }
            ccol++;
        }
        return false;
    }

    /**
     * Function to check topright to botleft diagonal with 1 chance in the center for min player
     * @return True if sequence is found, false otherwise.
     */
    private boolean toprightbotleft_1g_h() {

        // Top right to bottem left diagonal, with one gap in between                   
        //          [x]                 [x]
        //       [_]                 [x]
        //    [x]         of      [_]
        // [x]'                 [x]'

        int ccol = 0;
        int crow = 0;
        Disc d;

        while (ccol < fwidth - 2) {
            crow = 0;
            while (crow <= maxcol - 3) {
                d = field.getDiscAt(ccol, crow);
                if (d != null && d.getOwner() != turn) {
                    d = field.getDiscAt(ccol + 3, crow + 3);
                    if (d != null && d.getOwner() != turn) {
                        d = field.getDiscAt(ccol + 1, crow + 1);
                        Disc e = field.getDiscAt(ccol + 2, crow + 2);

                        if ((d != null && d.getOwner() != turn && e == null)
                                || (d == null && e != null && e.getOwner() != turn)) {
                            return true;
                        }
                    }
                }
                crow++;
            }
            ccol++;
        }
        return false;
    }

    /**
     * Function to check topleft to botright diagonal with 1 chance in the center for max player
     * @return True if sequence is found, false otherwise.
     */
    private boolean topleftbotright_1g_c() {

        // Top left to bottem right diagonal, with one gap in between                   
        // [x]                   [x]
        //   [_]                   [x]
        //      [x]         of       [_]
        //         [x]'                 [x]'

        int ccol = 3;
        int crow = 0;
        Disc d;

        while (ccol <= fwidth) {
            crow = 0;
            while (crow <= maxcol - 3) {
                d = field.getDiscAt(ccol, crow);
                if (d != null && d.getOwner() == turn) {
                    d = field.getDiscAt(ccol - 3, crow + 3);
                    if (d != null && d.getOwner() == turn) {
                        d = field.getDiscAt(ccol - 1, crow + 1);
                        Disc e = field.getDiscAt(ccol - 2, crow + 2);

                        if ((d != null && d.getOwner() == turn && e == null)
                                || (d == null && e != null && e.getOwner() == turn)) {
                            return true;
                        }
                    }
                }
                crow++;
            }
            ccol++;
        }
        return false;
    }

    /**
     * Function to check topleft to botright diagonal with 1 chance in the center for min player
     * @return True if sequence is found, false otherwise.
     */
    private boolean topleftbotright_1g_h() {

        // Top left to bottem right diagonal, with one gap in between                   
        // [x]                   [x]
        //   [_]                   [x]
        //      [x]         of       [_]
        //         [x]'                 [x]'
        int ccol = 3;
        int crow = 0;
        Disc d;

        while (ccol <= fwidth) {
            crow = 0;
            while (crow <= maxcol - 3) {
                d = field.getDiscAt(ccol, crow);
                if (d != null && d.getOwner() != turn) {
                    d = field.getDiscAt(ccol - 3, crow + 3);
                    if (d != null && d.getOwner() != turn) {
                        d = field.getDiscAt(ccol - 1, crow + 1);
                        Disc e = field.getDiscAt(ccol - 2, crow + 2);

                        if ((d != null && d.getOwner() != turn && e == null)
                                || (d == null && e != null && e.getOwner() != turn)) {
                            return true;
                        }
                    }
                }
                crow++;
            }
            ccol++;
        }
        return false;
    }

    /**
     * Function to check horizontal with 1 chance on either end for max player
     * @return True if sequence is found, false otherwise.
     */
    private boolean horizontal_1_c() {

        // Horizontal three in a row, one empty side
        // [_] [x] [x] [x]' of [x] [x] [x]' [_]
        int col = 2;
        int row = 0;
        int cons = 0;
        while (row <= maxcol) {
            Disc d = field.getDiscAt(col, row);
            if (d != null && d.getOwner() == turn) {
                cons++;
            } else {
                cons = 0;
            }

            if (cons == 3) {
                if (col - 2 == 0) { // left side
                    if (field.getDiscAt(col + 1, row) == null) {
                        return true;
                    }
                } else {
                    if (col == fwidth) { // most right side
                        if (field.getDiscAt(col - 3, row) == null) {
                            return true;
                        }
                    } else {
                        if (field.getDiscAt(col - 3, row) == null || field.getDiscAt(col + 1, row) == null) {
                            return true;
                        }
                    }
                }
            }
            col++;
            if (col == fwidth + 1) {
                row++;
                col = 0;
            }
        }
        return false;
    }

    /**
     * Function to check horizontal with 1 chance on either end for min player
     * @return True if sequence is found, false otherwise.
     */
    private boolean horizontal_1_h() {
        // Horizontal three in a row, one empty side
        // [_]' [x] [x] [x] of [x]' [x] [x] [_]
        int col = 2;
        int row = 0;
        int cons = 0;
        while (row <= maxcol) {
            Disc d = field.getDiscAt(col, row);
            if (d != null && d.getOwner() != turn) {
                cons++;
            } else {
                cons = 0;
            }

            if (cons == 3) {
                if (col - 2 == 0) { // left side
                    if (field.getDiscAt(col + 1, row) == null) {
                        return true;
                    }
                } else {
                    if (col == fwidth) { // most right side
                        if (field.getDiscAt(col - 3, row) == null) {
                            return true;
                        }
                    } else {
                        if (field.getDiscAt(col - 3, row) == null || field.getDiscAt(col + 1, row) == null) {
                            return true;
                        }
                    }
                }
            }
            col++;
            if (col == fwidth + 1) {
                row++;
                col = 0;
            }
        }
        return false;
    }

    /**
     * Function to check virtual 3 in a row with chance for max player
     * @return True if sequence is found, false otherwise.
     */
    private boolean vertical_1_c() {
        // Vertical three in a row, with one empty place at top
        // [_]
        // [x]
        // [x]
        // [x]'
        int col = 0;
        int row = 0;
        int cons = 0;
        while (col <= fwidth) {
            if (field.discsincol[col] >= 3) {
                Disc d = field.getDiscAt(col, row);
                if (d != null && d.getOwner() == turn) {
                    cons++;
                } else {
                    cons = 0;
                }
                row++;
                if (cons == 3 && field.getDiscAt(col, row) == null) {
                    return true;
                }
                if (row >= fheight) {
                    col++;
                    row = 0;
                }
            } else {
                col++;
            }
        }
        return false;
    }

    /**
     * Function to check virtual 3 in a row with chance for min player
     * @return True if sequence is found, false otherwise.
     */
    private boolean vertical_1_h() {

        // Vertical three in a row, with one empty place at top
        // [_]
        // [x]
        // [x]
        // [x]'
        int col = 0;
        int row = 0;
        int cons = 0;
        while (col <= fwidth) {
            if (field.discsincol[col] >= 3) {
                Disc d = field.getDiscAt(col, row);
                if (d != null && d.getOwner() != turn) {
                    cons++;
                } else {
                    cons = 0;
                }
                row++;
                if (cons == 3 && field.getDiscAt(col, row) == null) {
                    return true;
                }

                if (row >= fheight) {
                    col++;
                    row = 0;
                }
            } else {
                col++;
            }
        }
        return false;
    }

    /**
     * Function to check horizontal with chance in center for max player
     * @return True if sequence is found, false otherwise.
     */
    private boolean horizontal_1gap_c() {

        // Horizontal four in a row, but with one gap
        // [x]' [_] [x] [x] or [x]' [x] [_] [x]
        int col = 0;
        int row = 0;
        while (row <= maxcol) {
            Disc d = field.getDiscAt(col, row);
            if (d != null && d.getOwner() == turn) {
                d = field.getDiscAt(col + 3, row);
                if (d != null && d.getOwner() == turn) {
                    d = field.getDiscAt(col + 1, row);
                    Disc e = field.getDiscAt(col + 2, row);
                    if ((d != null && d.getOwner() == turn
                            && e == null) || (e != null && e.getOwner() == turn
                            && d == null)) {
                        return true;
                    }
                }

            }
            col++;
            if (col >= fwidth - 2) {
                col = 0;
                row++;
            }
        }
        return false;
    }

    /**
     * Function to check horizontal with chance in center for max player
     * @return True if sequence is found, false otherwise.
     */
    private boolean horizontal_1gap_h() {

        // Horizontal four in a row, but with one gap
        // [x]' [_] [x] [x] or [x]' [x] [_] [x]

        int col = 0;
        int row = 0;
        while (row <= maxcol) {
            Disc d = field.getDiscAt(col, row);
            if (d != null && d.getOwner() != turn) {
                d = field.getDiscAt(col + 3, row);
                if (d != null && d.getOwner() != turn) {
                    d = field.getDiscAt(col + 1, row);
                    Disc e = field.getDiscAt(col + 2, row);
                    if ((d != null && d.getOwner() != turn
                            && e == null) || (e != null && e.getOwner() != turn
                            && d == null)) {
                        return true;
                    }
                }

            }
            col++;
            if (col >= fwidth - 2) {
                col = 0;
                row++;
            }
        }
        return false;
    }

    /**
     * Function to check virtual 2 in a row with chance for max player
     * @return True if sequence is found, false otherwise.
     */
    private boolean vertical_2_c() {
        // Vertical two in a row, with one empty place at top
        // [_]
        // [x]
        // [x]'
        int col = 0;
        int row = 0;
        int cons = 0;
        while (col <= fwidth) {
            if (field.discsincol[col] >= 2) {
                Disc d = field.getDiscAt(col, row);
                if (d != null && d.getOwner() == turn) {
                    cons++;
                } else {
                    cons = 0;
                }
                if (cons == 2 && field.getDiscAt(col, row + 1) == null) {
                    return true;
                }
                row++;
                if (row >= fheight) {
                    col++;
                    row = 0;
                }
            } else {
                col++;
            }
        }
        return false;
    }

    /**
     * Function to check virtual 2 in a row with chance for min player
     * @return True if sequence is found, false otherwise.
     */
    private boolean vertical_2_h() {
        // Vertical two in a row, with one empty place at top
        // [_]
        // [x]
        // [x]'
        int col = 0;
        int row = 0;
        int cons = 0;
        while (col <= fwidth) {
            if (field.discsincol[col] >= 2) {
                Disc d = field.getDiscAt(col, row);
                if (d != null && d.getOwner() == turn) {
                    cons++;
                } else {
                    cons = 0;
                }
                if (cons == 2 && field.getDiscAt(col, row + 1) == null) {
                    return true;
                }
                row++;
                if (row >= fheight) {
                    col++;
                    row = 0;
                }
            } else {
                col++;
            }
        }
        return false;
    }

    /**
     * Function to check horizontal combinations with 2 discs for both players
     * @return Value of the most important combination, 1 if no such combination present.
     */
    private int horizontaltwo_2gap_c() {


        // Combinations of two horizontal disks
        // Checks for human player as well as computer player
        // Checks;
        // [_][x][_][x][_]
        // [x][_][x][_] or [_][x][_][x]
        // [x][_][_][x]
        // [x][X][_][_] or [_][_][X][x] or [_][x][x][_]

        int[] binNr = new int[maxcol + 1];
        int row = 0;
        int col = 0;
        int line = 0;

        while (row <= maxcol) {
            line = 0;
            col = 0;
            while (col <= fwidth) {
                Disc d = field.getDiscAt(col, row);
                if (d != null) {
                    if (d.getOwner() == turn) {
                        line = line << 2 | 3; //11
                    } else {
                        line = line << 2 | 2; //10, disc belongs to opponent
                    }
                } else {
                    line = line << 2 | 0; //00
                }
                col++;
            }
            binNr[row] = line;
            int pattern = 0xCC; //0011001100
            int mask = 0x3FF; //1111111111
            for (int j = 0; j <= fwidth; j++) {
                if ((line & mask) == pattern) {
                    return 400; //[_][x][_][x][_]
                }
                pattern <<= 2;
                mask <<= 2;
            }

            row++;
        }

        for (int i = 0; i <= maxcol; i++) {
            line = binNr[i];
            int pattern1 = 0xCC; //11001100
            int pattern2 = 0x33; //00110011
            int mask = 0xFF; //11111111
            for (int j = 0; j <= fwidth; j++) {
                if ((line & mask) == pattern1 || (line & mask) == pattern2) {
                    return 390; //[x][_][x][_] or [_][x][_][x]
                }
                pattern1 <<= 2;
                pattern2 <<= 2;
                mask <<= 2;
            }
        }
        for (int i = 0; i <= maxcol; i++) {
            line = binNr[i];
            int pattern1 = 0x88; //10001000
            int pattern2 = 0x22; //00100010
            int mask = 0xFF; //11111111
            for (int j = 0; j <= fwidth; j++) {
                if ((line & mask) == pattern1 || (line & mask) == pattern2) {
                    return -390; //[x][_][x][_] or [_][x][_][x]
                }
                pattern1 <<= 2;
                pattern2 <<= 2;
                mask <<= 2;
            }
        }
        for (int i = 0; i <= maxcol; i++) {
            line = binNr[i];
            int pattern = 0xC3; //11000011
            int mask = 0xFF; //11111111
            for (int j = 0; j <= fwidth; j++) {
                if ((line & mask) == pattern) {
                    return 380; //[x][_][_][x]
                }
                pattern <<= 2;
                mask <<= 2;
            }
        }
        for (int i = 0; i <= maxcol; i++) {
            line = binNr[i];
            int pattern = 0x82; //10000010
            int mask = 0xFF; //11111111
            for (int j = 0; j <= fwidth; j++) {
                if ((line & mask) == pattern) {
                    return -380; //[x][_][_][x]
                }
                pattern <<= 2;
                mask <<= 2;
            }
        }
        for (int i = 0; i <= maxcol; i++) {
            line = binNr[i];
            int pattern1 = 0xF0; //11110000
            int pattern2 = 0x0F; //00001111
            int pattern3 = 0x3C; //00111100
            int mask = 0xFF; //11111111
            for (int j = 0; j <= fwidth; j++) {
                if ((line & mask) == pattern1 || (line & mask) == pattern2 || (line & mask) == pattern3) {
                    return 350; //[x][X][_][_] or [_][_][X][x] or [_][x][x][_]
                }
                pattern1 <<= 2;
                pattern2 <<= 2;
                pattern3 <<= 2;
                mask <<= 2;
            }
        }
        for (int i = 0; i <= maxcol; i++) {
            line = binNr[i];
            int pattern1 = 0xA0; //10100000
            int pattern2 = 0x0A; //00001010
            int pattern3 = 0x28; //00101000
            int mask = 0xFF; //11111111
            for (int j = 0; j <= fwidth; j++) {
                if ((line & mask) == pattern1 || (line & mask) == pattern2 || (line & mask) == pattern3) {
                    return -350; //[x][X][_][_] or [_][_][X][x] or [_][x][x][_]
                }
                pattern1 <<= 2;
                pattern2 <<= 2;
                pattern3 <<= 2;
                mask <<= 2;
            }
        }

        return 1;
    }

    //Unused function, no documentation
//    private void horizontal_sequence(int[] sequences) {
//        final int w = field.getWidth(); //field dimensions
//        final int h = maxcol + 1; //check only the lower part of the field that actually contains discs
//        int maxseqlen1 = 0, maxseqlen2 = 0; //records the longest sequence for both players
//        Player p1 = turn;
//        for (int y = 0; y < maxcol; y++) {
//            int seqlen1 = 0, seqlen2 = 0; //length of the current sequence
//            boolean seqempty1 = false, seqempty2 = false; //whether an empty slot is next to the sequence
//            for (int x = 0; x < w; x++) {
//                Disc d = field.getDiscAt(x, y);
//                if (d == null) {
//                    if (seqlen1 > maxseqlen1) {
//                        maxseqlen1 = seqlen1;
//                    }
//                    if (seqlen2 > maxseqlen2) {
//                        maxseqlen2 = seqlen2;
//                    }
//                    seqempty1 = seqempty2 = true;
//                    seqlen1 = seqlen2 = 0;
//                } else if (d.getOwner() == p1) {
//                    if (seqempty2 && seqlen2 > maxseqlen2) {
//                        maxseqlen2 = seqlen2;
//                    }
//                    seqlen1++;
//                    seqlen2 = 0;
//                } else {
//                    if (seqempty1 && seqlen1 > maxseqlen1) {
//                        maxseqlen1 = seqlen1;
//                    }
//                    seqlen1 = 0;
//                    seqlen2++;
//                }
//            }
//            if (seqempty1 && seqlen1 > maxseqlen1) {
//                maxseqlen1 = seqlen1;
//            }
//            if (seqempty2 && seqlen2 > maxseqlen2) {
//                maxseqlen2 = seqlen2;
//            }
//        }
//        sequences[0] = maxseqlen1;
//        sequences[1] = maxseqlen2;
//    }
}
