/**
 * Creates the topnode of the tree.
 * The tree is used to find the best move using the minimaxAB method to optimize the result.
 * - Detailed design minimaxAB
 */
package AI;

import Game.PlayingField;

/**
 *
 * @author Twan Braamse/Ruben Jonk
 */
public class Tree {

    public Node topNode;// The top node of the tree.

    /**
     * Create a node that acts as the root of a new tree. 
     * @param diff The difficulty of the computer player that this tree is used for.
     * @param afield The playing field that represents the current field of the game. 
     */
    public Tree(int diff, PlayingField afield) {
        topNode = new Node(diff * 2, null, afield, -1);
    }

    /**
     * The initialization of the minimax algorithm.
     * 
     * The method uses the difficulty of the computer player to determine the 
     * depth of the search tree. The values of each possible moves are calculated, 
     * and stored in an array. The best move is stored. If there are multiple best 
     * moves, one of these moves is the next move.
     * @param diff The difficulty of the computer player.
     * @return The column of the best move.
     */
    public int minimaxAB(int diff) {

        int[] movelist = new int[topNode.field.getWidth()];
        int bmove = 0;  //store the best move
        int bmoveammount = 1;  //keeps track of the amount of best moves

        for (int i = 0; i < topNode.field.getWidth(); i++) {
            movelist[i] = alpha_beta_pruning(topNode.setChild(), diff * 2 - 1, -1000, 1000, false);
            System.out.println("move " + i + " has value " + movelist[i]);
            //For every move, check if the move is better than the current best move.
            //if true: assign the move to the best move and reset bmoveammount to 1.
            //if false: do nothing
            if (movelist[i] > movelist[bmove]) {
                bmove = i;
                bmoveammount = 1;
            }

            //For every ove, cmheck if the move is equal to the current best move
            //if true: compare if a random, uniformly generated number is smaller than the given predicate*.
            //      if true, assign the new move as best move and increment bmoveammount.
            //if false: do nothing
            //
            //*The predicate is 1 / #best moves. For n best moves, the chance to get the first
            //best move becomes 1/2 * 2/3 ... (n-1)/n = 1/n.
            if (movelist[i] == movelist[bmove]) {
                if (Math.random() < 1.0 / (bmoveammount + 1)) {
                    bmove = i;
                }
                bmoveammount++;
            }
        }

        return bmove;
    }

    /**
     * The alpha beta and minimax combination. 
     * 
     * @param node A node in the tree. Contains function to determine it's heuristic value.
     * @param depth The depth in the tree. depth == 0 indicates the node is a leaf.
     * @param a Alpha
     * @param b Beta
     * @param isComp Keeps track of max or minplayer.
     * @return The heuristic value of a node, either from a leaf or derived from a leaf by recursion.
     */
    public int alpha_beta_pruning(Node node, int depth, int a, int b, boolean isComp) {

        if (depth <= 0 || node.heurval >= 999 || node.heurval <= -999 || node.field.isFull()) {
            int heurval;//System.out.println("heuristic value: " + node.calHeuristicValue());
            heurval = node.calHeuristicValue();
            //if (node.heurval != 0 && node.heurval != 1) System.out.println("Heurval node = " + node.heurval);
            return heurval;
        }

        if (isComp) {  //  isComp == maxplayer

            while (node.nchilds < node.field.getWidth()) {
                a = max(a, alpha_beta_pruning(node.setChild(), depth - 1, a, b, !isComp));
                if (b <= a) {
                    break;
                }
            }
            return a;
        } else { //if (!isComp) == minplayer

            while (node.nchilds < node.field.getWidth()) {

                b = min(b, alpha_beta_pruning(node.setChild(), depth - 1, a, b, !isComp));
                if (b <= a) {
                    break;
                }
            }
            return b;
        }
    }

    /**
     * Returns the maximum of two integers.
     * @param x An integer.
     * @param y An integer.
     * @return The maximum of x and y.
     */
    private int max(int x, int y) {
        if (x > y) {
            return x;
        } else {
            return y;
        }

    }

    /**
     * Returns the minimum of two integers.
     * @param x An integer.
     * @param y An integer.
     * @return The minimum of x and y.
     */
    private int min(int x, int y) {
        if (x < y) {
            return x;
        } else {
            return y;
        }
    }
}
