package GUI;

import ScoreManagement.*;
import java.awt.Font;
import java.text.DecimalFormat;
import java.util.Arrays;
import javax.swing.DefaultComboBoxModel;
import javax.swing.JLabel;


/**
 * Handles displaying of high scores
 */
public class HighScore extends javax.swing.JDialog {
    /**
     * Creates new HighScore window
     * @param parent The parent window
     * @param modal Whether to show this window as a modal window
     */
    public HighScore(java.awt.Frame parent, boolean modal) {
        super(parent, modal);
        initComponents();
        fillUserLists();
        displayScoreAI();
    }

    /* Fill/update the user dropdown lists */
    private void fillUserLists() {
        jComboP1.setModel(new DefaultComboBoxModel(ScoreManagement.getHumanUsers()));
        jComboP2.setModel(new DefaultComboBoxModel(ScoreManagement.getHumanUsers()));
        comboDiff.setModel(new DefaultComboBoxModel(ScoreManagement.getComputerUsers()));
    }

    /* Internal class that stores calculated human-versus-computer scores */
    private class ScoreAI implements Comparable<ScoreAI> {
        int wins;
        int nofGames;
        int moves;
        User user;

	/**
	 * Create an instance associated with the given computer user
         * Pre: int ID is a valid User identifier
         * Post: New ScoreAI is made, with the id of the user
	 */
        ScoreAI(User usr) {
            user = usr;
        }

	/* Compares this instance to another instance for sorting */
        @Override
        public int compareTo (ScoreAI comp) {
            if (this.getWinPercentage() < comp.getWinPercentage()) { return 1; }
            if (this.getWinPercentage() > comp.getWinPercentage()) { return -1;  }
            
            // Nothing is returned, players have the same winning
            // percentage. Now we're going to look at the number of
            // total moves
            
            if (this.getMoves() < comp.getMoves()) { return 1; }
            if (this.getMoves() > comp.getMoves()) { return -1; }
            
            return 0;
        }

	/* Increment number of wins */
        public void addWin() {
            wins++;
        }

	/* Increment number of games */
        public void addGame() {
            nofGames++;
        }

	/* Add moves to counter */
        public void addMoves(int m) {
            moves += m;
        }

	/* Get the winning percentage */
        double getWinPercentage() {
            return ((double)wins / nofGames) * 100;
        }

	/* Get the total number of moves */
        int getMoves() {
            return moves;
        }

	/* Generates a string representation of this instance to display */
        @Override
        public String toString() {
            DecimalFormat df = new DecimalFormat("#.##");
            return user.toString()+" ("+df.format(this.getWinPercentage())+"%)";
        }
    }

    /**
     * Calculates and sorts the per-user score versus the selected computer user
     * Pre:     The users has selected an item in comboDiff
     *          The argument contains the array that has to be considered
     * Post:    The scores for AI are saved into array and returned
     * @param gameScores The list of game results to process
     * @return Sorted list of per-user scores
     */
    private ScoreAI[] calculateScoreAI(GameResult[] gameScores) {
        // First, determine the level and save it as Computerplayer
        User computer = (User)comboDiff.getSelectedItem();

        // How many game results are there?
        int noGameResults = gameScores.length;
        int count = 0;
        // Make a new array, selection of games which have to be considered
        ScoreAI selOfScores[] = new ScoreAI[noGameResults];

        for (GameResult g : gameScores) {
            // Check if the GameResult is the result of a game with the computer
            // and the corresponding difficulty level
            if (g.player1 == computer || g.player2 == computer) {
                // Determine the human player
                User hPlayer = (g.player1 == computer) ? g.player2 : g.player1;
                ScoreAI updateRecord = null;

                // Check if the user already exists in temporary database
                for (ScoreAI s : selOfScores) {
                    if (s != null && s.user == hPlayer) {
                        updateRecord = s;
                    }
                }
                
                // No such user found, create a new one
                if (updateRecord == null) {
                    selOfScores[count] = new ScoreAI(hPlayer);
                    updateRecord = selOfScores[count];
                    count++;
                }
                
                // Update the record with the wins
                updateRecord.addGame();
                updateRecord.addMoves(g.moves);
                if (g.winner == hPlayer) { updateRecord.addWin(); }

            }

        }
        // The array is too big, resize it to -count-
        ScoreAI scoreRecord[] = new ScoreAI[count];
        for (int i = 0; i != count; i++) {
            scoreRecord[i] = selOfScores[i];
        }
        
        // Now order it
        Arrays.sort(scoreRecord);
        // And display it
        return scoreRecord;

    }

    /* Display/update the human versus AI score for the selected difficulty level */
    private void displayScoreAI() {
        panelLabels.removeAll();
        
        ScoreAI[] allTime = calculateScoreAI(ScoreManagement.games);
        ScoreAI[] lastWeek = calculateScoreAI(ScoreManagement.getGamesWeek());
        ScoreAI[] lastDay = calculateScoreAI(ScoreManagement.getGamesDay());
        
        JLabel c1 = new JLabel();
        JLabel c2 = new JLabel("Last Day");
        JLabel c3 = new JLabel("Last week");
        JLabel c4 = new JLabel("All time");

        Font f = c2.getFont();
        c2.setFont(f.deriveFont(f.getStyle() ^ Font.BOLD));
        c3.setFont(f.deriveFont(f.getStyle() ^ Font.BOLD));
        c4.setFont(f.deriveFont(f.getStyle() ^ Font.BOLD));

        panelLabels.add(c1);
        panelLabels.add(c2);
        panelLabels.add(c3);
        panelLabels.add(c4);        
        
        for (int i = 0; i != 10; i++) {
            JLabel label1 = new JLabel(Integer.toString(i + 1));
            label1.setHorizontalAlignment(JLabel.CENTER);
            
            String strLabel2, strLabel3, strLabel4;
            
            if (i < lastDay.length) {
                strLabel2 = lastDay[i].toString();
            } else {
                strLabel2 = "";
            }
            JLabel label2 = new JLabel(strLabel2);
            
            if (i < lastWeek.length) {
                strLabel3 = lastWeek[i].toString();
            } else {
                strLabel3 = "";
            }
            JLabel label3 = new JLabel(strLabel3);
            
            if (i < allTime.length) {
                strLabel4 = allTime[i].toString();
            } else {
                strLabel4 = "";
            }
            JLabel label4 = new JLabel(strLabel4);
            
            panelLabels.add(label1);
            panelLabels.add(label2);
            panelLabels.add(label3);
            panelLabels.add(label4);
        }
	panelLabels.doLayout();
    }

    /* Calculate and display player-vs-player score for the selected players */
    private void calculateScorePlayer() {
        // First, determine the players
        User Player1 = (User) jComboP1.getSelectedItem();
        User Player2 = (User) jComboP2.getSelectedItem();

        // Init the integers that keep track of the total wins
        int winP1 = 0;
        int winP2 = 0;

        for (int i = 0; i != ScoreManagement.games.length; i++) {
            // Check if the game contains both Players in arbitrary order
            if ((ScoreManagement.games[i].player1 == Player1
                    && ScoreManagement.games[i].player2 == Player2) || (ScoreManagement.games[i].player1 == Player2
                    && ScoreManagement.games[i].player2 == Player1)) {

                // Check winner and increase sumber
                if (ScoreManagement.games[i].winner == Player1) {
                    winP1++;
                }
                if (ScoreManagement.games[i].winner == Player2) {
                    winP2++;
                }
            }
        }

        // We have the scores, display them in the GUI
        wonPlayer1.setText(Integer.toString(winP1));
        wonPlayer2.setText(Integer.toString(winP2));
    }

    /** This method is called from within the constructor to
     * initialize the form.
     * WARNING: Do NOT modify this code. The content of this method is
     * always regenerated by the Form Editor.
     */
    @SuppressWarnings("unchecked")
    // <editor-fold defaultstate="collapsed" desc="Generated Code">//GEN-BEGIN:initComponents
    private void initComponents() {

        jPanel1 = new javax.swing.JPanel();
        jLabel1 = new javax.swing.JLabel();
        jPanel2 = new javax.swing.JPanel();
        jPanelHuman = new javax.swing.JPanel();
        jLabel3 = new javax.swing.JLabel();
        jComboP1 = new javax.swing.JComboBox();
        jComboP2 = new javax.swing.JComboBox();
        wonPlayer1 = new javax.swing.JLabel();
        wonPlayer2 = new javax.swing.JLabel();
        jPanelComputer = new javax.swing.JPanel();
        jLabel6 = new javax.swing.JLabel();
        comboDiff = new javax.swing.JComboBox();
        panelLabels = new javax.swing.JPanel();

        setDefaultCloseOperation(javax.swing.WindowConstants.DISPOSE_ON_CLOSE);
        setTitle("Highscores");
        setResizable(false);

        jPanel1.setBackground(new java.awt.Color(52, 127, 191));
        jPanel1.setMinimumSize(new java.awt.Dimension(200, 500));
        jPanel1.setName("jPanel1"); // NOI18N

        jLabel1.setFont(new java.awt.Font("Tahoma", 1, 24));
        jLabel1.setForeground(new java.awt.Color(255, 255, 255));
        jLabel1.setText("Highscores");
        jLabel1.setName("jLabel1"); // NOI18N

        jPanel2.setName("jPanel2"); // NOI18N

        jPanelHuman.setBackground(new java.awt.Color(220, 220, 220));
        jPanelHuman.setName("jPanelHuman"); // NOI18N

        jLabel3.setFont(new java.awt.Font("Tahoma", 1, 14));
        jLabel3.setText("Versus Human");
        jLabel3.setName("jLabel3"); // NOI18N

        jComboP1.setName("jComboP1"); // NOI18N
        jComboP1.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                jComboP1ActionPerformed(evt);
            }
        });

        jComboP2.setName("jComboP2"); // NOI18N
        jComboP2.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                jComboP2ActionPerformed(evt);
            }
        });

        wonPlayer1.setBackground(new java.awt.Color(255, 255, 255));
        wonPlayer1.setFont(new java.awt.Font("Tahoma", 0, 24));
        wonPlayer1.setText("0");
        wonPlayer1.setName("wonPlayer1"); // NOI18N

        wonPlayer2.setFont(new java.awt.Font("Tahoma", 0, 24));
        wonPlayer2.setText("0");
        wonPlayer2.setName("wonPlayer2"); // NOI18N

        javax.swing.GroupLayout jPanelHumanLayout = new javax.swing.GroupLayout(jPanelHuman);
        jPanelHuman.setLayout(jPanelHumanLayout);
        jPanelHumanLayout.setHorizontalGroup(
            jPanelHumanLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGroup(javax.swing.GroupLayout.Alignment.TRAILING, jPanelHumanLayout.createSequentialGroup()
                .addContainerGap(401, Short.MAX_VALUE)
                .addComponent(jLabel3)
                .addContainerGap())
            .addGroup(jPanelHumanLayout.createSequentialGroup()
                .addGroup(jPanelHumanLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING, false)
                    .addGroup(jPanelHumanLayout.createSequentialGroup()
                        .addContainerGap()
                        .addComponent(jComboP1, javax.swing.GroupLayout.PREFERRED_SIZE, 99, javax.swing.GroupLayout.PREFERRED_SIZE))
                    .addGroup(jPanelHumanLayout.createSequentialGroup()
                        .addGap(56, 56, 56)
                        .addComponent(wonPlayer1, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, Short.MAX_VALUE)))
                .addGroup(jPanelHumanLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING, false)
                    .addGroup(jPanelHumanLayout.createSequentialGroup()
                        .addGap(12, 12, 12)
                        .addComponent(jComboP2, javax.swing.GroupLayout.PREFERRED_SIZE, 99, javax.swing.GroupLayout.PREFERRED_SIZE))
                    .addGroup(jPanelHumanLayout.createSequentialGroup()
                        .addGap(56, 56, 56)
                        .addComponent(wonPlayer2, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, Short.MAX_VALUE)))
                .addGap(104, 104, 104))
        );
        jPanelHumanLayout.setVerticalGroup(
            jPanelHumanLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGroup(jPanelHumanLayout.createSequentialGroup()
                .addContainerGap()
                .addComponent(jLabel3)
                .addGap(18, 18, 18)
                .addGroup(jPanelHumanLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.BASELINE)
                    .addComponent(jComboP1, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE)
                    .addComponent(jComboP2, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE))
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.UNRELATED)
                .addGroup(jPanelHumanLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.BASELINE)
                    .addComponent(wonPlayer2)
                    .addComponent(wonPlayer1))
                .addContainerGap(javax.swing.GroupLayout.DEFAULT_SIZE, Short.MAX_VALUE))
        );

        jPanelComputer.setBackground(new java.awt.Color(220, 220, 220));
        jPanelComputer.setName("jPanelComputer"); // NOI18N

        jLabel6.setFont(new java.awt.Font("Tahoma", 1, 14));
        jLabel6.setText("Versus Computer");
        jLabel6.setName("jLabel6"); // NOI18N

        comboDiff.setName("comboDiff"); // NOI18N
        comboDiff.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                comboDiffActionPerformed(evt);
            }
        });

        panelLabels.setName("panelLabels"); // NOI18N
        panelLabels.setLayout(new java.awt.GridLayout(0, 4));

        javax.swing.GroupLayout jPanelComputerLayout = new javax.swing.GroupLayout(jPanelComputer);
        jPanelComputer.setLayout(jPanelComputerLayout);
        jPanelComputerLayout.setHorizontalGroup(
            jPanelComputerLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGroup(jPanelComputerLayout.createSequentialGroup()
                .addContainerGap()
                .addGroup(jPanelComputerLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
                    .addComponent(panelLabels, javax.swing.GroupLayout.DEFAULT_SIZE, 492, Short.MAX_VALUE)
                    .addGroup(jPanelComputerLayout.createSequentialGroup()
                        .addComponent(comboDiff, javax.swing.GroupLayout.PREFERRED_SIZE, 123, javax.swing.GroupLayout.PREFERRED_SIZE)
                        .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED, 248, Short.MAX_VALUE)
                        .addComponent(jLabel6)))
                .addContainerGap())
        );
        jPanelComputerLayout.setVerticalGroup(
            jPanelComputerLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGroup(jPanelComputerLayout.createSequentialGroup()
                .addContainerGap()
                .addGroup(jPanelComputerLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.BASELINE)
                    .addComponent(jLabel6)
                    .addComponent(comboDiff, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE))
                .addGap(18, 18, 18)
                .addComponent(panelLabels, javax.swing.GroupLayout.PREFERRED_SIZE, 295, javax.swing.GroupLayout.PREFERRED_SIZE)
                .addContainerGap())
        );

        javax.swing.GroupLayout jPanel2Layout = new javax.swing.GroupLayout(jPanel2);
        jPanel2.setLayout(jPanel2Layout);
        jPanel2Layout.setHorizontalGroup(
            jPanel2Layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGroup(javax.swing.GroupLayout.Alignment.TRAILING, jPanel2Layout.createSequentialGroup()
                .addContainerGap()
                .addGroup(jPanel2Layout.createParallelGroup(javax.swing.GroupLayout.Alignment.TRAILING)
                    .addComponent(jPanelComputer, javax.swing.GroupLayout.Alignment.LEADING, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, Short.MAX_VALUE)
                    .addComponent(jPanelHuman, javax.swing.GroupLayout.Alignment.LEADING, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, Short.MAX_VALUE))
                .addContainerGap())
        );
        jPanel2Layout.setVerticalGroup(
            jPanel2Layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGroup(jPanel2Layout.createSequentialGroup()
                .addContainerGap()
                .addComponent(jPanelHuman, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE)
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                .addComponent(jPanelComputer, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, Short.MAX_VALUE)
                .addContainerGap())
        );

        javax.swing.GroupLayout jPanel1Layout = new javax.swing.GroupLayout(jPanel1);
        jPanel1.setLayout(jPanel1Layout);
        jPanel1Layout.setHorizontalGroup(
            jPanel1Layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGroup(jPanel1Layout.createSequentialGroup()
                .addContainerGap()
                .addGroup(jPanel1Layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
                    .addComponent(jPanel2, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, Short.MAX_VALUE)
                    .addComponent(jLabel1, javax.swing.GroupLayout.Alignment.TRAILING))
                .addContainerGap())
        );
        jPanel1Layout.setVerticalGroup(
            jPanel1Layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGroup(jPanel1Layout.createSequentialGroup()
                .addContainerGap()
                .addComponent(jLabel1)
                .addGap(11, 11, 11)
                .addComponent(jPanel2, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE)
                .addGap(38, 38, 38))
        );

        getContentPane().add(jPanel1, java.awt.BorderLayout.CENTER);

        pack();
    }// </editor-fold>//GEN-END:initComponents

    /* PvP Player 1 selection has changed, update score */
private void jComboP1ActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_jComboP1ActionPerformed
    calculateScorePlayer();
}//GEN-LAST:event_jComboP1ActionPerformed

    /* PvP Player 2 selection has changed, update score */
private void jComboP2ActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_jComboP2ActionPerformed
    calculateScorePlayer();
}//GEN-LAST:event_jComboP2ActionPerformed

    /* PvAI difficulty level has changed, update score */
private void comboDiffActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_comboDiffActionPerformed
    displayScoreAI();
}//GEN-LAST:event_comboDiffActionPerformed

    // Variables declaration - do not modify//GEN-BEGIN:variables
    private javax.swing.JComboBox comboDiff;
    private javax.swing.JComboBox jComboP1;
    private javax.swing.JComboBox jComboP2;
    private javax.swing.JLabel jLabel1;
    private javax.swing.JLabel jLabel3;
    private javax.swing.JLabel jLabel6;
    private javax.swing.JPanel jPanel1;
    private javax.swing.JPanel jPanel2;
    private javax.swing.JPanel jPanelComputer;
    private javax.swing.JPanel jPanelHuman;
    private javax.swing.JPanel panelLabels;
    private javax.swing.JLabel wonPlayer1;
    private javax.swing.JLabel wonPlayer2;
    // End of variables declaration//GEN-END:variables
}
